from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_allclose, assert_array_equal

setup_logging()

@MPITest([1, 4])
def test_paint(comm):

    from pmesh.pm import ParticleMesh

    # initialize a random white noise field in real space
    pm = ParticleMesh(Nmesh=(8, 8, 8), BoxSize=(128, 128, 128.), comm=comm)
    real = pm.generate_whitenoise(type='real', seed=3333) # a RealField

    # FFT to a ComplexField
    complex = real.r2c()

    # gather to all ranks --> shape is now (8,8,8) on all ranks
    data = numpy.concatenate(comm.allgather(numpy.array(real.ravel())))\
                 .reshape(real.cshape)

    # mesh from data array
    realmesh = ArrayMesh(data, BoxSize=128., comm=comm)

    # paint() must yield the RealField
    assert_array_equal(real, realmesh.to_field(mode='real'))

    # gather complex to all ranks --> shape is (8,8,5) on all ranks
    cdata = numpy.concatenate(comm.allgather(numpy.array(complex.ravel())))\
                 .reshape(complex.cshape)

    # mesh from complex array
    cmesh = ArrayMesh(cdata, BoxSize=128., comm=comm)

    # paint() yields the ComplexField
    assert_allclose(complex, cmesh.to_field(mode='complex'), atol=1e-8)



from nbodykit.base.mesh import MeshSource
from nbodykit import CurrentMPIComm, mockmaker
from nbodykit.utils import attrs_to_dict
import numpy

class LinearMesh(MeshSource):
    """
    A MeshSource object that generates a :class:`~pmesh.pm.RealField` density
    mesh from a linear power spectrum function :math:`P(k)`.

    Parameters
    ----------
    Plin: callable
        the callable linear power spectrum function, which takes the
        wavenumber as its single argument
    BoxSize : float, 3-vector of floats
        the size of the box to generate the grid on
    Nmesh : int, 3-vector of int
        the number of the mesh cells per side
    seed : int, optional
        the global random seed, used to set the seeds across all ranks
    remove_variance : bool, optional
        .. deprecated:: 0.2.9
            use ``unitary_amplitude`` instead
    unitary_amplitude: bool, optional
        ``True`` to remove variance from the complex field by fixing the
        amplitude to :math:`P(k)` and only the phase is random.
    inverted_phase: bool, optional
        ``True`` to invert phase of the complex field by fixing the
        amplitude to :math:`P(k)` and only the phase is random.
    comm : MPI communicator
        the MPI communicator
    """
    def __repr__(self):
        return "LinearMesh(seed=%(seed)d)" % self.attrs

    @CurrentMPIComm.enable
    def __init__(self, Plin, BoxSize, Nmesh, seed=None,
            unitary_amplitude=False,
            inverted_phase=False,
            remove_variance=None,
            comm=None):

        self.Plin = Plin

        # cosmology and communicator
        self.comm    = comm

        self.attrs.update(attrs_to_dict(Plin, 'plin.'))

        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)
        self.attrs['seed'] = seed
        if remove_variance is not None:
            unitary_amplitude = remove_variance

        self.attrs['unitary_amplitude'] = unitary_amplitude
        self.attrs['inverted_phase'] = inverted_phase

        MeshSource.__init__(self, BoxSize=BoxSize, Nmesh=Nmesh, dtype='f4', comm=comm)

    def to_complex_field(self):
        """
        Return a ComplexField, generating from the linear power spectrum.

        .. note::

            The density field is normalized to :math:`1+\delta` such that
            the mean of the return field in real space is unity.

        Returns
        -------
        :class:`pmesh.pm.ComplexField`
            an array-like object holding the generated linear density
            field in Fourier space
        """
        # generate linear density field with desired seed
        complex, _ = mockmaker.gaussian_complex_fields(self.pm, self.Plin, self.attrs['seed'],
                    unitary_amplitude=self.attrs['unitary_amplitude'],
                    inverted_phase=self.attrs['inverted_phase'],
                    compute_displacement=False)
        # set normalization to 1 + \delta.
        def filter(k, v):
            mask = numpy.bitwise_and.reduce([ki == 0 for ki in k])
            v[mask] = 1.0
            return v

        complex.apply(filter)
        complex.attrs = {}
        complex.attrs.update(self.attrs)
        return complex



from nbodykit.cosmology import Planck15, Cosmology
from nbodykit.cosmology.background import RadiationDominated, MatterDominated
from numpy.testing import assert_allclose
import numpy

def test_radiation_dominated():

    C = Planck15
    a = numpy.logspace(-2, 0, 11, endpoint=True)
    z = 1 / a - 1

    pt = RadiationDominated(C, a=a)

    """
    from scipy.optimize import minimize
    def f(a_H):
        pt.a_H = a_H
        pt._D1, pt._D2 = pt._solve()
        D1 = pt.D1(a)
        D_CC = C.scale_independent_growth_factor(z)
        return ((D1 / D_CC - 1) ** 2).sum() ** 0.5
    a_Hbest = minimize(f, pt.a_H, method='Nelder-Mead', options=dict(xatol=1e-7, fatol=1e-7)).x
    print(a_Hbest)
    pt.a_H = a_Hbest
    """
    # linear growth function
    D1 = pt.D1(a)
    D_CC = C.scale_independent_growth_factor(z)
    assert_allclose(D1, D_CC, rtol=1e-3)

    # linear growth rate
    f1 = pt.f1(a)
    f_CC = C.scale_independent_growth_rate(z)
    assert_allclose(f1, f_CC, rtol=1e-3)

    # second order quantities
    D2 = pt.D2(a)
    f2 = pt.f2(a)

def test_matter_dominated():
    C = Planck15.clone(T_cmb=0.1, m_ncdm=[]) # suppress cmb and remove neutrino
    a = numpy.logspace(-2, 0, 11, endpoint=True)
    z = 1 / a - 1

    pt = MatterDominated(C.Omega0_m, a=a)

    # linear growth function
    D1 = pt.D1(a)
    D_CC = C.scale_independent_growth_factor(z)
    assert_allclose(D1, D_CC, rtol=1e-3)

    # linear growth rate
    f1 = pt.f1(a)
    f_CC = C.scale_independent_growth_rate(z)
    assert_allclose(f1, f_CC, rtol=1e-3)

    # second order quantities
    D2 = pt.D2(a)
    f2 = pt.f2(a)



from nbodykit.source.catalog import HaloCatalog, HDFCatalog
from nbodykit import CurrentMPIComm, transform
from nbodykit.cosmology import Cosmology

class DemoHaloCatalog(HaloCatalog):
    """
    Create a demo catalog of halos using one of the built-in :mod:`halotools`
    catalogs.

    .. note::
        The first request for a particular catalog will download the data
        and cache in the ``~/.astropy/cache/halotools`` directory.

    Parameters
    ----------
    simname : string
        Nickname of the simulation. Currently supported simulations are
        Bolshoi  (simname = ``bolshoi``), Consuelo (simname = ``consuelo``),
        MultiDark (simname = ``multidark``), and Bolshoi-Planck (simname = ``bolplanck``).
    halo_finder : string
        Nickname of the halo-finder, e.g. ``rockstar`` or ``bdm``.
    redshift : float
        Redshift of the requested snapshot.
        Must match one of the available snapshots within ``dz_tol=0.1``,
        or a prompt will be issued providing the nearest
        available snapshots to choose from.

    Examples
    --------
    >>> from nbodykit.tutorials import DemoHaloCatalog
    >>> halos = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)
    >>> print(halos.columns)
    """
    @CurrentMPIComm.enable
    def __init__(self, simname, halo_finder, redshift, comm=None):

        from halotools.sim_manager import CachedHaloCatalog, DownloadManager
        from halotools.sim_manager.supported_sims import supported_sim_dict

        # do seme setup
        self.comm = comm
        meta_cols = ['Lbox', 'redshift', 'particle_mass']

        # try to automatically load from the Halotools cache
        exception = None
        if self.comm.rank == 0:
            kws = {'simname':simname, 'halo_finder':halo_finder, 'redshift':redshift}
            try:
                cached_halos = CachedHaloCatalog(dz_tol=0.1, **kws)
                fname = cached_halos.fname # the filename to load
                meta = {k:getattr(cached_halos, k) for k in meta_cols}
            except Exception as e:

                # try to download on the root rank
                try:
                    # download
                    dl = DownloadManager()
                    dl.download_processed_halo_table(dz_tol=0.1, **kws)

                    # access the cached halo catalog and get fname attribute
                    # NOTE: this does not read the data
                    cached_halos = CachedHaloCatalog(dz_tol=0.1, **kws)
                    fname = cached_halos.fname
                    meta = {k:getattr(cached_halos, k) for k in meta_cols}
                except Exception as e:
                    exception = e
        else:
            fname = None
            meta = None

        # re-raise a download error on all ranks if it occurred
        exception = self.comm.bcast(exception, root=0)
        if exception is not None:
            raise exception

        # broadcast the file we are loading
        fname = self.comm.bcast(fname, root=0)
        meta = self.comm.bcast(meta, root=0)

        # initialize an HDF catalog and add Position/Velocity
        cat = HDFCatalog(fname, comm=comm)
        cat['Position'] = transform.StackColumns(cat['halo_x'], cat['halo_y'], cat['halo_z'])
        cat['Velocity'] = transform.StackColumns(cat['halo_vx'], cat['halo_vy'], cat['halo_vz'])

        # get the cosmology from Halotools
        cosmo = supported_sim_dict[simname]().cosmology # this is astropy cosmology
        cosmo = Cosmology.from_astropy(cosmo)

        # initialize the HaloCatalog
        HaloCatalog.__init__(self, cat, cosmo, meta['redshift'], mdef='vir', mass='halo_mvir')

        # add some meta-data
        # NOTE: all Halotools catalogs have to these attributes
        self.attrs['BoxSize'] = meta['Lbox']
        self.attrs['redshift'] = meta['redshift']
        self.attrs['particle_mass'] = meta['particle_mass']

        # save the cosmology
        self.cosmo = cosmo
        self.attrs['cosmo'] = dict(self.cosmo)



import numpy

def bin_ndarray(ndarray, new_shape, weights=None, operation=numpy.mean):
    """
    Bins an ndarray in all axes based on the target shape, by summing
    or averaging.

    Parameters
    ----------
    ndarray : array_like
        the input array to re-bin
    new_shape : tuple
        the tuple holding the desired new shape
    weights : array_like, optional
        weights to multiply the input array by, before running the re-binning
        operation,

    Notes
    -----
    *   Dimensions in `new_shape` must be integral factor smaller than the
        old shape
    *   Number of output dimensions must match number of input dimensions.
    *   See https://gist.github.com/derricw/95eab740e1b08b78c03f

    Examples
    --------
    >>> m = numpy.arange(0,100,1).reshape((10,10))
    >>> n = bin_ndarray(m, new_shape=(5,5), operation=numpy.sum)
    >>> print(n)
    [[ 22  30  38  46  54]
     [102 110 118 126 134]
     [182 190 198 206 214]
     [262 270 278 286 294]
     [342 350 358 366 374]]
    """
    if ndarray.shape == new_shape:
        raise ValueError("why are we re-binning if the new shape equals the old shape?")
    if ndarray.ndim != len(new_shape):
        raise ValueError("Shape mismatch: {} -> {}".format(ndarray.shape, new_shape))
    if numpy.sometrue(numpy.mod(ndarray.shape, new_shape)):
        args = (str(new_shape), str(ndarray.shape))
        msg = "desired shape of %s must be integer factor smaller than the old shape %s" %args
        raise ValueError(msg)

    compression_pairs = [(d, c//d) for d, c in zip(new_shape, ndarray.shape)]
    flattened = [l for p in compression_pairs for l in p]
    ndarray = ndarray.reshape(flattened)
    if weights is not None: weights = weights.reshape(flattened)

    for i in range(len(new_shape)):
        if weights is not None:
            ndarray = operation(ndarray*weights, axis=-1*(i+1))
            weights = numpy.sum(weights, axis=-1*(i+1))
            ndarray /= weights
        else:
            ndarray = operation(ndarray, axis=-1*(i+1))
    return ndarray

class BinnedStatistic(object):
    """
    Lightweight class to hold statistics binned at fixed coordinates.

    For example, this class could hold a grid of (r, mu) or (k, mu) bins
    for a correlation function or power spectrum measurement.

    It is modeled after the syntax of :class:`xarray.Dataset`, and is designed
    to hold correlation function or power spectrum results (in 1D or 2D)

    Parameters
    ----------
    dims : list, (Ndim,)
        A list of strings specifying names for the coordinate dimensions.
        The dimension names stored in :attr:`dims` have the suffix 'cen'
        added, to indicate that the coordinate grid is defined at the bin
        centers
    edges : list, (Ndim,)
        A list specifying the bin edges for each dimension
    data : array_like
        a structured array holding the data variables, where the named
        fields interpreted as the variable names. The variable names are
        stored in :attr:`variables`
    fields_to_sum : list, optional
        the name of fields that will be summed when reindexing, instead
        of averaging
    **kwargs :
        Any additional keywords are saved as metadata in the :attr:`attrs`
        dictionary attribute

    Examples
    --------
    The following example shows how to read a power spectrum
    measurement from a JSON file, as output by nbodykit, assuming
    the JSON file holds a dictionary with a 'power' entry holding the
    relevant data

    >>> filename = 'test_data.json'
    >>> pk = BinnedStatistic.from_json(['k'], filename, 'power')

    In older versions of nbodykit, results were written using plaintext ASCII
    files. Although now deprecated, this type of files can be read using:

    >>> filename = 'test_data.dat'
    >>> dset = BinnedStatistic.from_plaintext(['k'], filename)

    Data variables can be accessed in a dict-like fashion:

    >>> power = pkmu['power'] # returns power data variable

    Array-like indexing of a :class:`BinnedStatistic` returns a new :class:`BinnedStatistic`
    holding the sliced data:

    >>> pkmu
    <BinnedStatistic: dims: (k: 200, mu: 5), variables: ('mu', 'k', 'power')>
    >>> pkmu[:,0] # select first mu column
    <BinnedStatistic: dims: (k: 200), variables: ('mu', 'k', 'power')>

    Additional data variables can be added to the :class:`BinnedStatistic` via:

    >>> modes = numpy.ones((200, 5))
    >>> pkmu['modes'] = modes

    Coordinate-based indexing is possible through :func:`sel`:

    >>> pkmu
    <BinnedStatistic: dims: (k: 200, mu: 5), variables: ('mu', 'k', 'power')>
    >>> pkmu.sel(k=slice(0.1, 0.4), mu=0.5)
    <BinnedStatistic: dims: (k: 30), variables: ('mu', 'k', 'power')>

    :func:`squeeze` will explicitly squeeze the specified dimension
    (of length one) such that the resulting instance has one less dimension:

    >>> pkmu
    <BinnedStatistic: dims: (k: 200, mu: 1), variables: ('mu', 'k', 'power')>
    >>> pkmu.squeeze(dim='mu') # can also just call pkmu.squeeze()
    <BinnedStatistic: dims: (k: 200), variables: ('mu', 'k', 'power')>

    :func:`average` returns a new :class:`BinnedStatistic` holding the
    data averaged over one dimension

    :func:`reindex` will re-bin the coordinate arrays along the specified
    dimension
    """
    def __init__(self, dims, edges, data, fields_to_sum=[], coords=None, **kwargs):

        # number of dimensions must match
        if len(dims) != len(edges):
            raise ValueError("size mismatch between specified `dims` and `edges`")

        # input data must be structured array
        if not isinstance(data, numpy.ndarray) or data.dtype.names is None:
            raise TypeError("'data' should be a structured numpy array")

        shape = tuple(len(e)-1 for e in edges)
        if data.shape != shape:
            args = (shape, data.shape)
            raise ValueError("`edges` imply data shape of %s, but data has shape %s" %args)

        self.dims = list(dims)
        self.edges = dict(zip(self.dims, edges))

        # coordinates are the bin centers
        self.coords = {}
        for i, dim in enumerate(self.dims):
            if coords is not None and coords[i] is not None:
                self.coords[dim] = numpy.copy(coords[i])
            else:
                self.coords[dim] = 0.5 * (edges[i][1:] + edges[i][:-1])

        # store variables as a structured array
        self.data = data.copy()

        # define a mask such that a coordinate grid element will be masked
        # if any of the variables at that coordinate are (NaN, inf)
        self.mask = numpy.zeros(self.shape, dtype=bool)
        for name in data.dtype.names:
            self.mask = numpy.logical_or(self.mask, ~numpy.isfinite(self.data[name]))

        # fields that we wish to sum, instead of averaging
        self._fields_to_sum = fields_to_sum

        # save and track metadata
        self.attrs = {}
        for k in kwargs: self.attrs[k] = kwargs[k]

    @classmethod
    def from_state(kls, state):
        obj = kls(dims=state['dims'], edges=state['edges'], coords=state['coords'], data=state['data'])
        obj.attrs.update(state['attrs'])
        return obj

    def __getstate__(self):
        return dict(
                dims=self.dims,
                edges=[self.edges[dim] for dim in self.dims],
                coords=[self.coords[dim] for dim in self.dims],
                data=self.data,
                attrs=self.attrs)

    @property
    def shape(self):
        """
        The shape of the coordinate grid
        """
        return tuple(len(self.coords[d]) for d in self.dims)

    @property
    def variables(self):
        """
        Alias to return the names of the variables stored in `data`
        """
        return list(self.data.dtype.names)

    @classmethod
    def __construct_direct__(cls, data, mask, **kwargs):
        """
        Shortcut around __init__ for internal use to construct and
        return a new class instance. The returned object should be
        identical to that returned by __init__.

        Notes
        -----
        *   Useful for returning new instances with sliced data/mask
        *   The keyword arguments required to create a full, unbroken
            instance are `dims`, `coords`, `edges`, and `attrs`

        Parameters
        ----------
        data :
        """
        obj = object.__new__(cls)
        for k in kwargs: setattr(obj, k, kwargs[k])

        for k, d in zip(['data', 'mask'], [data, mask]):
            setattr(obj, k, d)
            if obj.shape != d.shape:
                try:
                    setattr(obj, k, d.reshape(obj.shape))
                except:
                    raise ValueError("shape mismatch between data and coordinates")
        return obj

    def __copy_attrs__(self):
        """
        Return a copy of all necessary attributes associated with
        the `BinnedStatistic`. This dictionary + `data` and `mask` are all
        that's required to reconstruct a new class
        """
        kw = {}
        kw['dims'] = list(self.dims)
        kw['edges'] = self.edges.copy()
        kw['coords'] = self.coords.copy()
        kw['attrs'] = self.attrs.copy()
        kw['_fields_to_sum'] = list(self._fields_to_sum)
        return kw

    def __finalize__(self, data, mask, indices):
        """
        Finalize and return a new instance from a slice of the
        current object (returns a copy)
        """
        edges, coords = self.__slice_edges__(indices)
        kw = {'dims':list(self.dims), 'edges':edges, 'coords':coords,
              'attrs':self.attrs.copy(), '_fields_to_sum':self._fields_to_sum}
        return self.__class__.__construct_direct__(data, mask, **kw)

    def __slice_edges__(self, indices):
        """
        Internal function to slice the `edges` attribute with the
        specified indices, which specify the included coordinate bins
        """
        edges = {}
        coords = {}
        for i, dim in enumerate(self.dims):
            if len(indices[i]) > 0:
                idx = list(indices[i]) + [indices[i][-1]+1]
            else:
                idx = [0]
            edges[dim] = self.edges[dim][idx]
            coords[dim] = 0.5 * (edges[dim][1:] + edges[dim][:-1])

        return edges, coords

    def __str__(self):
        name = self.__class__.__name__
        dims = "(" + ", ".join(['%s: %d' %(k, self.shape[i]) for i, k in enumerate(self.dims)]) + ")"

        if len(self.variables) < 5:
            return "<%s: dims: %s, variables: %s>" %(name, dims, str(tuple(self.variables)))
        else:
            return "<%s: dims: %s, variables: %d total>" %(name, dims, len(self.variables))

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return iter(self.variables)

    def __contains__(self, key):
        return key in self.variables

    def __setitem__(self, key, data):
        """
        Add a new variable with the name `key` to the class using `data`
        """
        if numpy.shape(data) != self.data.shape:
            raise ValueError("data to be added must have shape %s" %str(self.data.shape))

        # add the new (key, type) to the dtype, or if the key is present, overwrite
        dtype = list(self.data.dtype.descr) # make copy
        names = list(self.data.dtype.names) # make copy
        if key in names:
            i = names.index(key)
            dtype.pop(i); names.pop(i)
        dtype += [(key, data.dtype.type)]

        # add old variables
        dtype = numpy.dtype(dtype)
        new = numpy.zeros(self.data.shape, dtype=dtype)
        for col in names:
            new[col] = self.data[col]

        # the new data to add
        new[key] = data
        mask = numpy.logical_or(self.mask, ~numpy.isfinite(new[key]))

        # save
        self.data = new
        self.mask = mask

    def __getitem__(self, key):
        """
        Index- or string- based indexing

        Notes
        -----
        *   If a single string is passed, the key is intrepreted
            as a `variable` or `coordinate`, and the corresponding
            array is returned
        *   If a list of strings is passed, then a new `BinnedStatistic`
            holding only the `variable` names in `key` is returned
        *   Integer-based indexing or slices similar to numpy
            indexing will slice `data`, returning a new
            `BinnedStatistic` holding the newly sliced data and coordinate grid
        *   Scalar indexes (i.e., integers) used to index a certain
            dimension will "squeeze" that dimension, removing it
            from the coordinate grid
        """
        # if single string passed, return a coordinate or variable
        if isinstance(key, str):
            if key in self.variables:
                return self.data[key]
            else:
                raise KeyError("`%s` is not a valid variable name" %key)

        # indices to slice the data with
        indices = [list(range(0, self.shape[i])) for i in range(len(self.dims))]

        # check for list/tuple of variable names
        # if so, return a BinnedStatistic with slice of columns
        if isinstance(key, (list, tuple)) and all(isinstance(x, str) for x in key):
            if all(k in self.variables for k in key):
                return self.__finalize__(self.data[list(key)], self.mask.copy(), indices)
            else:
                invalid = ', '.join("'%s'" %k for k in key if k not in self.variables)
                raise KeyError("cannot slice variables -- invalid names: (%s)" %invalid)

        key_ = key

        # if key is a single integer or single slice, make it a list
        make_iterable = isinstance(key, slice) or isinstance(key, int)
        if make_iterable or isinstance(key, list) and all(isinstance(x, int) for x in key):
            key_ = [key]

        squeezed_dims = []
        for i, subkey in enumerate(key_):
            if i >= len(self.dims):
                raise IndexError("too many indices for BinnedStatistic; note that ndim = %d" %len(self.dims))

            if isinstance(subkey, int):
                indices[i] = [subkey]
                squeezed_dims.append(self.dims[i])
            elif isinstance(subkey, list):
                indices[i] = subkey
            elif isinstance(subkey, slice):
                indices[i] = list(range(*subkey.indices(self.shape[i])))

        # can't squeeze all dimensions!!
        if len(squeezed_dims) == len(self.dims):
            raise IndexError("cannot return object with all remaining dimensions squeezed")

        # fail nicely if we fail at all
        try:
            toret = self.__finalize__(self.data[key], self.mask[key], indices)
            for dim in squeezed_dims:
                toret = toret.squeeze(dim)
            return toret
        except ValueError:
            raise IndexError("this type of slicing not implemented")

    def _get_index(self, dim, val, method=None):
        """
        Internal function to compute the bin index of the nearest
        coordinate value to the input value
        """
        index = self.coords[dim]
        if method == 'nearest':
            i = (numpy.abs(index-val)).argmin()
        else:
            try:
                i = list(index).index(val)
            except Exception as e:
                args = (dim, str(e))
                msg = "error converting '%s' index; try setting `method = 'nearest'`: %s"
                raise IndexError(msg %args)

        return i

    #--------------------------------------------------------------------------
    # user-called functions
    #--------------------------------------------------------------------------
    def to_json(self, filename):
        """
        Write a BinnedStatistic from a JSON file.

        .. note::
            This uses :class:`nbodykit.utils.JSONEncoder` to write the
            JSON file

        Parameters
        ----------
        filename : str
            the name of the file to write
        """
        import json
        from nbodykit.utils import JSONEncoder
        state = self.__getstate__()
        with open(filename, 'w') as ff:
            json.dump(state, ff, cls=JSONEncoder)

    @classmethod
    def from_json(cls, filename, key='data', dims=None, edges=None, **kwargs):
        """
        Initialize a BinnedStatistic from a JSON file.

        The JSON file should contain a dictionary, where the data to load is stored
        as the ``key`` entry, with an ``edges`` entry specifying bin edges, and
        optionally, a ``attrs`` entry giving a dict of meta-data

        .. note::
            This uses :class:`nbodykit.utils.JSONDecoder` to load the
            JSON file

        Parameters
        ----------
        filename : str
            the name of the file to load
        key : str, optional
            the name of the key in the JSON file holding the data to load
        dims : list, optional
            list of names specifying the dimensions, i.e., ``['k']`` or ``['k', 'mu']``;
            must be supplied if not given in the JSON file

        Returns
        -------
        dset : BinnedStatistic
            the BinnedStatistic holding the data from file
        """
        import json
        from nbodykit.utils import JSONDecoder

        # parse
        with open(filename, 'r') as ff:
            state = json.load(ff, cls=JSONDecoder)

        # the data
        if key not in state:
            args = (key, tuple(state.keys()))
            raise ValueError("no data entry found in JSON format for '%s' key; valid keys are %s" %args)
        data = state[key]

        # the dimensions
        dims = state.pop('dims', dims)
        if dims is None:
            raise ValueError("no `dims` found in JSON file; please specify as keyword argument")

        # the edges
        edges = state.pop('edges', edges)
        if edges is None:
            raise ValueError("no `edges` found in JSON file; please specify as keyword argument")

        # the coords
        coords = state.pop('coords', None)

        # meta-data
        attrs = state.pop('attrs', {})
        attrs.update(kwargs)

        return cls(dims, edges, data, coords=coords, **attrs)

    @classmethod
    def from_plaintext(cls, dims, filename, **kwargs):
        """
        Initialize a BinnedStatistic from a plaintext file

        .. note:: Deprecated in nbodykit 0.2.x
            Storage of BinnedStatistic objects as plaintext ASCII files is no longer supported;
            See :func:`BinnedStatistic.from_json`

        Parameters
        ----------
        dims : list
            list of names specifying the dimensions, i.e., ``['k']`` or ``['k', 'mu']``
        filename : str
            the name of the file to load

        Returns
        -------
        dset : BinnedStatistic
            the BinnedStatistic holding the data from file
        """
        import warnings
        msg = "storage of BinnedStatistic objects as ASCII plaintext files is deprecated; see BinnedStatistic.from_json"
        warnings.warn(msg, FutureWarning, stacklevel=2)

        # make sure dims is a list/tuple
        if not isinstance(dims, (tuple, list)):
            raise TypeError("`dims` should be a list or tuple of strings")

        # read from file
        try:
            if len(dims) == 1:
                data, meta = _Read1DPlainText(filename)
            elif len(dims) == 2:
                data, meta = _Read2DPlainText(filename)
        except Exception as e:
            msg = "unable to read plaintext file, perhaps the dimension of the file does "
            msg += "not match the passed `dims`;\nexception: %s" %str(e)
            raise ValueError(msg)

        # get the bin edges
        edges = meta.pop('edges', None)
        if edges is None:
            raise ValueError("plaintext file does not include `edges`; cannot be loaded into a BinnedStatistic")
        if len(dims) == 1:
            edges = [edges]

        meta.update(kwargs)
        return cls(dims, edges, data, **meta)


    def copy(self, cls=None):
        """
        Returns a copy of the BinnedStatistic, optionally change the type
        to cls. cls must be a subclass of BinnedStatistic.
        """
        attrs = self.__copy_attrs__()
        if cls is None:
            cls = self.__class__
        if not issubclass(cls, BinnedStatistic):
            raise TypeError("The cls argument must be a subclass of BinnedStatistic")

        return cls.__construct_direct__(self.data.copy(), self.mask.copy(), **attrs)

    def rename_variable(self, old_name, new_name):
        """
        Rename a variable in :attr:`data` from ``old_name`` to ``new_name``.

        Note that this procedure is performed in-place (does not
        return a new BinnedStatistic)

        Parameters
        ----------
        old_name : str
            the name of the old varibale to rename
        new_name : str
            the desired new variable name

        Raises
        ------
        ValueError
            If `old_name` is not present in :attr:`variables`
        """
        import copy
        if old_name not in self.variables:
            raise ValueError("`%s` is not an existing variable name" %old_name)
        new_dtype = copy.deepcopy(self.data.dtype)

        names = list(new_dtype.names)
        names[names.index(old_name)] = new_name
        new_dtype.names = names
        self.data.dtype = new_dtype

    def sel(self, method=None, **indexers):
        """
        Return a new BinnedStatistic indexed by coordinate values along the
        specified dimension(s).

        Notes
        -----
        Scalar values used to index a specific dimension will result in
        that dimension being squeezed. To keep a dimension of unit length,
        use a list to index (see examples below).

        Parameters
        ----------
        method : {None, 'nearest'}
            The method to use for inexact matches; if set to `None`, require
            an exact coordinate match, otherwise match the nearest coordinate
        **indexers :
            the pairs of dimension name and coordinate value used to index
            the BinnedStatistic

        Returns
        -------
        sliced : BinnedStatistic
            a new BinnedStatistic holding the sliced data and coordinate grid

        Examples
        --------
        >>> pkmu
        <BinnedStatistic: dims: (k: 200, mu: 5), variables: ('mu', 'k', 'power')>

        >>> pkmu.sel(k=0.4)
        <BinnedStatistic: dims: (mu: 5), variables: ('mu', 'k', 'power')>

        >>> pkmu.sel(k=[0.4])
        <BinnedStatistic: dims: (k: 1, mu: 5), variables: ('mu', 'k', 'power')>

        >>> pkmu.sel(k=slice(0.1, 0.4), mu=0.5)
        <BinnedStatistic: dims: (k: 30), variables: ('mu', 'k', 'power')>
        """
        indices = {}
        squeezed_dims = []
        for dim, key in indexers.items():

            if isinstance(key, list):
                indices[dim] = [self._get_index(dim, k, method=method) for k in key]
            elif isinstance(key, slice):
                new_slice = []
                for name in ['start', 'stop']:
                    new_slice.append(self._get_index(dim, getattr(key, name), method=method))
                i = self.dims.index(dim)
                indices[dim] = list(range(*slice(*new_slice).indices(self.shape[i])))
            elif not numpy.isscalar(key):
                raise IndexError("please index using a list, slice, or scalar value")
            else:
                indices[dim] = [self._get_index(dim, key, method=method)]
                squeezed_dims.append(dim)

        # can't squeeze all dimensions!!
        if len(squeezed_dims) == len(self.dims):
            raise IndexError("cannot return object with all remaining dimensions squeezed")

        toret = self.take(**indices)

        for dim in squeezed_dims:
            toret = toret.squeeze(dim)
        return toret

    def take(self, *masks, **indices):
        """
        Take a subset of a BinnedStatistic from given list of indices.
        This is more powerful but more verbose than `sel`. Also the
        result is never squeezed, even if only a single item along the direction
        is used.

        Parameters
        ----------
        masks : array_like (boolean)
            a list of masks that are of the same shape as the data.

        indices: dict (string : array_like)
            mapping from axes (by name, dim) to items to select (list/array_like).
            Each item is a valid selector for numpy's fancy indexing.

        Returns
        -------
        new BinnedStatistic, where only items selected by all axes are kept.

        Examples
        --------
        >>> pkmu
        <BinnedStatistic: dims: (k: 200, mu: 5), variables: ('mu', 'k', 'power')>

        # similar to pkmu.sel(k > 0.4), select the bin centers
        >>> pkmu.take(k=pkmu.coords['k'] > 0.4)
        <BinnedStatistic: dims: (mu: 5), variables: ('mu', 'k', 'power')>

        # also similar to pkmu.sel(k > 0.4), select the bin averages
        >>> pkmu.take(pkmu['k'] > 0.4)
        <BinnedStatistic: dims: (k: 30), variables: ('mu', 'k', 'power')>

        # impossible with sel.
        >>> pkmu.take(pkmu['modes'] > 0)

        """
        indices_dict = {}
        indices_dict.update(indices) # rename for a cleaner API.

        # flatten the masks, will keep items that are true everywhere
        mask = numpy.ones(self.shape, dtype='?')
        for m in masks: mask = mask & m

        indices = [numpy.ones(self.shape[i], dtype='?') for i in range(len(self.dims))]

        # update indices with masks
        for i, dim in enumerate(self.dims):
            axis = list(range(len(self.dims)))
            axis.remove(i)
            axis = tuple(axis)
            mask1 = mask.all(axis=axis)
            indices[i] &= mask1

        # update indices with indices_dict
        for dim, index in indices_dict.items():
            i = self.dims.index(dim)
            if isinstance(index, numpy.ndarray) and index.dtype == numpy.dtype('?'):
                # boolean mask?
                assert index.ndim == 1
                indices[i] &= index
            else:
                mask1 = numpy.zeros(self.shape[i], dtype='?')
                mask1.put(index, True)
                indices[i] &= mask1

        # convert to indices
        for i in range(len(indices)):
            indices[i] = indices[i].nonzero()[0]

        data = self.data.copy()
        mask = self.mask.copy()
        for i, idx in enumerate(indices):
            data = numpy.take(data, idx, axis=i)
            mask = numpy.take(mask, idx, axis=i)

        toret = self.__finalize__(data, mask, indices)

        return toret

    def squeeze(self, dim=None):
        """
        Squeeze the BinnedStatistic along the specified dimension, which
        removes that dimension from the BinnedStatistic.

        The behavior is similar to that of :func:`numpy.squeeze`.

        Parameters
        ----------
        dim : str, optional
            The name of the dimension to squeeze. If no dimension
            is provided, then the one dimension with unit length will
            be squeezed

        Returns
        -------
        squeezed : BinnedStatistic
            a new BinnedStatistic instance, squeezed along one dimension

        Raises
        ------
        ValueError
            If the specified dimension does not have length one, or
            no dimension is specified and multiple dimensions have
            length one

        Examples
        --------
        >>> pkmu
        <BinnedStatistic: dims: (k: 200, mu: 1), variables: ('mu', 'k', 'power')>
        >>> pkmu.squeeze() # squeeze the mu dimension
        <BinnedStatistic: dims: (k: 200), variables: ('mu', 'k', 'power')>
        """
        # infer the right dimension to squeeze
        if dim is None:
            dim = [k for k in self.dims if len(self.coords[k]) == 1]
            if not len(dim):
                raise ValueError("no available dimensions with length one to squeeze")
            if len(dim) > 1:
                raise ValueError("multiple dimensions available to squeeze -- please specify")
            dim = dim[0]
        else:
            if dim not in self.dims:
                raise ValueError("`%s` is not a valid dimension name" %dim)
            if len(self.coords[dim]) != 1:
                raise ValueError("the `%s` dimension must have length one to squeeze" %dim)

        # remove the dimension from the grid
        i = self.dims.index(dim)
        toret = self.copy()
        toret.dims.pop(i); toret.edges.pop(dim); toret.coords.pop(dim)
        if not len(toret.dims):
            raise ValueError("cannot squeeze the only remaining axis")

        # construct new object with squeezed data/mask
        attrs = toret.__copy_attrs__()
        d = toret.data.squeeze(axis=i)
        m = toret.mask.squeeze(axis=i)
        return self.__construct_direct__(d, m, **attrs)

    def average(self, dim, **kwargs):
        """
        Compute the average of each variable over the specified dimension.

        Parameters
        ----------
        dim : str
            The name of the dimension to average over
        **kwargs :
            Additional keywords to pass to :func:`BinnedStatistic.reindex`. See the
            documentation for :func:`BinnedStatistic.reindex` for valid keywords.

        Returns
        -------
        averaged : BinnedStatistic
            A new BinnedStatistic, with data averaged along one dimension,
            which reduces the number of dimension by one
        """
        spacing = (self.edges[dim][-1] - self.edges[dim][0])
        toret = self.reindex(dim, spacing, **kwargs)
        return toret.sel(**{dim:toret.coords[dim][0]})

    def reindex(self,
                    dim,
                    spacing,
                    weights=None,
                    force=True,
                    return_spacing=False,
                    fields_to_sum=[]):
        """
        Reindex the dimension ``dim`` by averaging over multiple coordinate bins,
        optionally weighting by ``weights``.

        Returns a new BinnedStatistic holding the re-binned data.

        Notes
        -----
        *   We can only re-bin to an integral factor of the current
            dimension size in order to inaccuracies when re-binning to
            overlapping bins
        *   Variables specified in `fields_to_sum` will
            be summed when re-indexing, instead of averaging


        Parameters
        ----------
        dim : str
            The name of the dimension to average over
        spacing : float
            The desired spacing for the re-binned data. If `force = True`,
            the spacing used will be the closest value to this value, such
            that the new bins are N times larger, when N is an integer
        weights : array_like or str, optional (`None`)
            An array to weight the data by before re-binning, or if
            a string is provided, the name of a data column to use as weights
        force : bool, optional
            If `True`, force the spacing to be a value such
            that the new bins are N times larger, when N is an integer,
            otherwise, raise an exception. Default is `True`
        return_spacing : bool, optional
            If `True`, return the new spacing as the second return value.
            Default is `False`.
        fields_to_sum : list
            the name of fields that will be summed when reindexing, instead
            of averaging

        Returns
        -------
        rebinned : BinnedStatistic
            A new BinnedStatistic instance, which holds the rebinned coordinate
            grid and data variables
        spacing : float, optional
            If `return_spacing` is `True`, the new coordinate spacing
            will be returned
        """
        i = self.dims.index(dim)
        fields_to_sum += self._fields_to_sum

        # determine the new binning
        old_spacings = numpy.diff(self.coords[dim])
        if not numpy.array_equal(old_spacings, old_spacings):
            raise ValueError("`reindex` requires even bin spacings")
        old_spacing = old_spacings[0]

        factor = numpy.round(spacing/old_spacing).astype('int')
        if not factor:
            raise ValueError("new spacing must be smaller than original spacing of %.2e" %old_spacing)
        if factor == 1:
            raise ValueError("closest binning size to input spacing is the same as current binning")
        if not numpy.allclose(old_spacing*factor, spacing) and not force:
            raise ValueError("if `force = False`, new bin spacing must be an integral factor smaller than original")

        # make a copy of the data
        data = self.data.copy()

        # get the weights
        if isinstance(weights, str):
            if weights not in self.variables:
                raise ValueError("cannot weight by `%s`; no such column" %weights)
            weights = self.data[weights]

        edges = self.edges[dim]
        new_shape = list(self.shape)

        # check if we need to discard bins from the end
        leftover = self.shape[i] % factor
        if leftover and not force:
            args = (leftover, old_spacing*factor)
            raise ValueError("cannot re-bin because they are %d extra bins, using spacing = %.2e" %args)
        if leftover:
            sl = [slice(None, None)]*len(self.dims)
            sl[i] = slice(None, -leftover)
            data = data[tuple(sl)]
            if weights is not None: weights = weights[sl]
            edges = edges[:-leftover]
            new_shape[i] = new_shape[i] - leftover

        # new edges
        new_shape[i] /= factor
        new_shape[i] = int(new_shape[i])
        new_edges = numpy.linspace(edges[0], edges[-1], new_shape[i]+1)

        # the re-binned data
        new_data = numpy.empty(new_shape, dtype=self.data.dtype)
        for name in self.variables:
            operation = numpy.nanmean
            weights_ = weights
            if weights is not None or name in fields_to_sum:
                operation = numpy.nansum
                if name in fields_to_sum: weights_ = None
            new_data[name] = bin_ndarray(data[name], new_shape, weights=weights_, operation=operation)

        # the new mask
        new_mask = numpy.zeros_like(new_data, dtype=bool)
        for name in self.variables:
            new_mask = numpy.logical_or(new_mask, ~numpy.isfinite(new_data[name]))

        # construct new object
        kw = self.__copy_attrs__()
        kw['edges'][dim] = new_edges
        kw['coords'][dim] = 0.5*(new_edges[1:] + new_edges[:-1])
        toret = self.__construct_direct__(new_data, new_mask, **kw)

        return (toret, spacing) if return_spacing else toret

#------------------------------------------------------------------------------
# Deprecated Plaintext read/write functions
#------------------------------------------------------------------------------
def _Read2DPlainText(filename):
    """
    Reads the plain text storage of a 2D measurement

    Returns
    -------
    data : dict
        dictionary holding the `edges` data, as well as the
        data columns for the P(k,mu) measurement
    metadata : dict
        any additional metadata to store as part of the
        P(k,mu) measurement
    """
    d = {}
    metadata = {}

    with open(filename, 'r') as ff:

        # read number of k and mu bins are first line
        Nk, Nmu = [int(l) for l in ff.readline().split()]
        N = Nk*Nmu

        # names of data columns on second line
        columns = ff.readline().split()

        lines = ff.readlines()
        data = numpy.array([float(l) for line in lines[:N] for l in line.split()])
        data = data.reshape((Nk, Nmu, -1)) #reshape properly to (Nk, Nmu)

        # make a dict, making complex arrays from real/imag parts
        i = 0
        while i < len(columns):
            name = columns[i]
            nextname = columns[i+1] if i < len(columns)-1 else ''
            if name.endswith('.real') and nextname.endswith('.imag'):
                name = name.split('.real')[0]
                d[name] = data[...,i] + 1j*data[...,i+1]
                i += 2
            else:
                d[name] = data[...,i]
                i += 1

        # store variables as a structured array
        dtypes = numpy.dtype([(name, d[name].dtype) for name in d])
        data = numpy.empty(data.shape[:2], dtype=dtypes)
        for name in d:
            data[name] = d[name]

        # read the edges for k and mu bins
        edges = []
        l1 = int(lines[N].split()[-1]); N = N+1
        edges.append(numpy.array([float(line) for line in lines[N:N+l1]]))
        l2 = int(lines[N+l1].split()[-1]); N = N+l1+1
        edges.append(numpy.array([float(line) for line in lines[N:N+l2]]))
        metadata['edges'] = edges

        # read any metadata
        if len(lines) > N+l2:
            N_meta = int(lines[N+l2].split()[-1])
            N = N + l2 + 1
            meta = lines[N:N+N_meta]
            for line in meta:
                fields = line.split()
                cast = fields[-1]
                if cast in __builtins__:
                    metadata[fields[0]] = __builtins__[cast](fields[1])
                elif hasattr(numpy, cast):
                     metadata[fields[0]] = getattr(numpy, cast)(fields[1])
                else:
                    raise TypeError("Metadata must have builtin or numpy type")

    return data, metadata

def _Read1DPlainText(filename):
    """
    Reads the plain text storage of a 1D measurement

    Notes
    -----
    *   If `edges` is present in the file, they will be returned
        as part of the metadata, with the key `edges`
    *   If the first line of the file specifies column names,
        they will be returned as part of the metadata with the
        `columns` key

    Returns
    -------
    data : array_like
        the 1D data stacked vertically, such that each columns
        represents a separate data variable
    metadata : dict
        any additional metadata to store as part of the
        P(k) measurement
    """
    # data list
    data = []

    # extract the metadata
    metadata = {}
    make_float = lambda x: float(x[1:])
    with open(filename, 'r') as ff:

        currline = 0
        lines = ff.readlines()

        # try to read columns
        if lines[0][0] == '#':
            try:
                metadata['columns'] = lines[0][1:].split()
            except:
                pass

        while True:

            # break if we are at the EOF
            if currline == len(lines): break
            line = lines[currline]

            if not line:
                currline += 1
                continue

            if line[0] != '#':
                data.append([float(l) for l in line.split()])
            else:
                line = line[1:]

                # read edges
                if 'edges' in line:
                    fields = line.split()
                    N = int(fields[-1]) # number of edges
                    metadata['edges'] = numpy.array([make_float(l) for l in lines[currline+1:currline+1+N]])
                    currline += 1+N
                    continue

                # read metadata
                if 'metadata' in line:
                    # read and cast the metadata properly
                    fields = line.split()
                    N = int(fields[-1]) # number of individual metadata lines
                    for i in range(N):
                        fields = lines[currline+1+i][1:].split()
                        cast = fields[-1]
                        if cast in __builtins__:
                            metadata[fields[0]] = __builtins__[cast](fields[1])
                        elif hasattr(numpy, cast):
                             metadata[fields[0]] = getattr(numpy, cast)(fields[1])
                        else:
                            raise TypeError("metadata must have builtin or numpy type")
                    currline += 1+N
                    continue

            # add to the data
            currline += 1

    data = numpy.asarray(data)

    # get names of columns, using default if not in file
    columns = metadata.pop('columns', ['col_%d' %i for i in range(data.shape[1])])

    # make a dict, making complex arrays from real/imag parts
    i = 0
    d = {}
    while i < len(columns):
        name = columns[i]
        nextname = columns[i+1] if i < len(columns)-1 else ''
        if name.endswith('.real') and nextname.endswith('.imag'):
            name = name.split('.real')[0]
            d[name] = data[...,i] + 1j*data[...,i+1]
            i += 2
        else:
            d[name] = data[...,i]
            i += 1

    # store variables as a structured array
    dtypes = numpy.dtype([(name, d[name].dtype) for name in d])
    data = numpy.empty(len(data), dtype=dtypes)
    for name in d:
        data[name] = d[name]

    return data, metadata



from nbodykit.binned_statistic import BinnedStatistic
import numpy
import warnings

class WedgeBinnedStatistic(BinnedStatistic):
    """
        A BinnedStatistic of wedges that can be converted to multiples.
    """

    def to_poles(self, poles):
        r"""
        Invert the measured wedges :math:`\xi(r,mu)` into correlation
        multipoles, :math:`\xi_\ell(r)`.

        To select a mu_range, use

        .. code:: python

            poles = self.sel(mu=slice(*mu_range), method='nearest').to_poles(poles)

        Parameters
        ----------
        poles: array_like
            the list of multipoles to compute

        Returns
        -------
        xi_ell : BinnedStatistic
            a data set holding the :math:`\xi_\ell(r)` multipoles
        """
        from scipy.special import legendre
        from scipy.integrate import quad

        # new data array
        x = str(self.dims[0])
        dtype = numpy.dtype([(x, 'f8')] + [('corr_%d' %ell, 'f8') for ell in poles])
        data = numpy.zeros((self.shape[0]), dtype=dtype)
        dims = [x]
        edges = [self.edges[x]]

        # FIXME: use something fancier than the central point.
        mu_bins = numpy.diff(self.edges['mu'])
        mu_mid = (self.edges['mu'][1:] + self.edges['mu'][:-1])/2.

        for ell in poles:
            legendrePolynomial = (2.*ell+1.)*legendre(ell)(mu_mid)
            data['corr_%d' %ell] = numpy.sum(self['corr']*legendrePolynomial*mu_bins,axis=-1)/numpy.sum(mu_bins)

        data[x] = numpy.mean(self[x],axis=-1)

        return BinnedStatistic(dims=dims, edges=edges, data=data, poles=poles)


class AnalyticUniformRandoms(object):
    """
    Internal class to compute analytic pair counts for uniformly
    distributed randoms.

    This can compute the expected randoms counts for several coordinate
    choices, based on ``mode``. The relevant volume/area calculations for the
    various coordinate modes are:

    * mode='1d': volume of sphere
    * mode='2d': volume of spherical sector
    * mode='projected': volume of cylinder
    * mode='angular': area of spherical cap
    """
    def __init__(self, mode, dims, edges, BoxSize):

        assert mode in ['1d', '2d', 'projected', 'angular']
        self.mode = mode
        self.edges = edges
        self.dims = dims
        self.BoxSize = BoxSize

    def get_filling_factor(self):
        """
        This gives the ratio of the volume (or area) occupied by each bin to
        the global volume (or area).

        It is different based on the value of :attr:`mode`.
        """
        # based on a volume of a sphere
        if self.mode == '1d':
            r_edges = self.edges['r']
            V = 4. / 3. * numpy.pi * r_edges**3
            dV = numpy.diff(V)
            return dV / self.BoxSize.prod()

        # based on the volume of a spherical sector
        elif self.mode == '2d':
            r_edges, mu_edges = self.edges['r'], self.edges['mu']
            V = 2. / 3. * numpy.pi * numpy.outer(r_edges**3, mu_edges)
            V *= 2.0 # accounts for the fact that we bin in abs(mu)
            dV = numpy.diff(numpy.diff(V, axis=0), axis=1)
            return  dV / self.BoxSize.prod()

        # based on volume of a cylinder
        elif self.mode == 'projected':
            rp_edges, pi_edges = self.edges['rp'], self.edges['pi']
            V = numpy.pi * numpy.outer(rp_edges**2, (2. * pi_edges)) # height is double pimax!
            dV = numpy.diff(numpy.diff(V, axis=0), axis=1)
            return  dV / self.BoxSize.prod()

        # based on the surface area of a spherical cap
        elif self.mode == 'angular':
            theta_bins = self.edges['theta']
            chord = 2*numpy.sin(0.5*numpy.deg2rad(theta_bins)) # chord distance
            h = 1. - numpy.sqrt(1. - chord**2)
            A = numpy.pi*(chord**2 + h**2)
            dA = numpy.diff(A)
            return dA / (4*numpy.pi)

    def __call__(self, NR1, NR2=None):
        """
        Evaluate the expected randoms pair counts, and the total_weight, returns
        as an object that looks like the result of paircount.

        """
        edges = [self.edges[d] for d in self.dims] # sequentialize it, poor API!

        if NR2 is None:
            R1R2 = NR1 ** 2  * self.get_filling_factor()
            total_wnpairs = NR1 * (NR1 - 1) * 0.5
        else:
            R1R2 = NR1 * NR2 * self.get_filling_factor()
            total_wnpairs = NR1 * NR2 * 0.5

        data = numpy.empty_like(R1R2, dtype=[('npairs', 'f8'), ('wnpairs', 'f8')])
        data['npairs'] = R1R2
        data['wnpairs'] = R1R2
        pairs = WedgeBinnedStatistic(self.dims, edges, data)

        R1R2 = lambda : None
        R1R2.attrs = {}
        R1R2.pairs = pairs
        R1R2.attrs['total_wnpairs']= total_wnpairs
        R1R2.pairs.attrs['total_wnpairs']= total_wnpairs

        return R1R2

def LandySzalayEstimator(pair_counter, data1, data2, randoms1, randoms2, R1R2=None, logger=None, **kwargs):
    """
    Compute the correlation function from data/randoms using the
    Landy - Szalay estimator to compute the correlation function.

    Parameters
    ----------
    pair_counter : SimulationBoxPairCount, SurveyDataPairCount
        the pair counting algorithm class
    data1 : CatalogSource
        the first data source
    data2 : CatalogSource, None
        the second data catalog to cross-correlate; can be None for auto-correlations
    randoms1 : CatalogSource
        the randoms catalog corresponding to ``data1``
    randoms2 : CatalogSource, None
        the second randoms catalog; can be None for auto-correlations
    R1R2 : SimulationBoxPairCount, SurveyDataPairCount, optional
        if provided, random pairs R1R2 are not recalculated
    **kwargs :
        the parameters passed to the ``pair_counter`` class to count pairs

    Returns
    -------
    D1D2, D1R2, D2R1, R1R2, CF : BinnedStatistic
        the various terms of the LS estimator + the correlation function result

    References
    ----------
    http://adsabs.harvard.edu/abs/1993ApJ...412...64L
    """
    # make sure we have the randoms
    assert randoms1 is not None
    comm = data1.comm

    if randoms2 is None: randoms2 = randoms1

    # and randoms - randoms calculation
    if logger is not None and comm.rank == 0:
        logger.info("computing randoms1 - randoms2 pair counts")
    if not R1R2:
        R1R2 = pair_counter(first=randoms1, second=randoms2, **kwargs)

    # data1 x data2
    if logger is not None and comm.rank == 0:
        logger.info("computing data1 - data2 pair counts")
    D1D2 = pair_counter(first=data1, second=data2, **kwargs)

    # do data - randoms correlation
    if logger is not None and comm.rank == 0:
        logger.info("computing data1 - randoms2 pair counts")
    D1R2 = pair_counter(first=data1, second=randoms2, **kwargs)

    if data2 is not None:
        if logger is not None and comm.rank == 0:
            logger.info("computing data2 - randoms1 pair counts")
        D2R1 = pair_counter(first=data2, second=randoms1, **kwargs)
    else:
        D2R1 = D1R2

    fDD = R1R2.attrs['total_wnpairs']/D1D2.attrs['total_wnpairs']
    fDR = R1R2.attrs['total_wnpairs']/D1R2.attrs['total_wnpairs']
    fRD = R1R2.attrs['total_wnpairs']/D2R1.attrs['total_wnpairs']

    nonzero = R1R2.pairs['npairs'] > 0

    # init
    CF = numpy.zeros(D1D2.pairs.shape)
    CF[:] = numpy.nan
    Error = numpy.zeros(D1D2.pairs.shape)
    Error[:] = numpy.nan

    # the Landy - Szalay estimator
    # (DD - DR - RD + RR) / RR
    DD = (D1D2.pairs['wnpairs'])[nonzero]
    DR = (D1R2.pairs['wnpairs'])[nonzero]
    RD = (D2R1.pairs['wnpairs'])[nonzero]
    RR = (R1R2.pairs['wnpairs'])[nonzero]
    xi = (fDD * DD - fDR * DR - fRD * RD)/RR + 1
    CF[nonzero] = xi[:]

    # warn about NaNs in the estimator
    if comm.rank == 0 and numpy.isnan(CF).any():
        msg = ("The RR calculation in the Landy-Szalay estimator contains"
        " separation bins with no bins. This will result in NaN values in the resulting"
        " correlation function. Try increasing the number of randoms and/or using"
        " broader bins.")
        warnings.warn(msg)

    CF = _create_tpcf_result(D1D2.pairs, CF)
    return D1D2.pairs, D1R2.pairs, D2R1.pairs, R1R2.pairs, CF

def NaturalEstimator(D1D2):
    """
    Internal function to computing the correlation function using
    analytic randoms and the so-called "natural" correlation function
    estimator, :math:`DD/RR - 1`.
    """
    attrs = D1D2.attrs

    # determine the sample sizes
    if attrs['is_cross']:
        ND1, ND2 = attrs['N1'], attrs['N2']
    else:
        ND1, ND2 = attrs['N1'], None

    mode = attrs['mode']
    BoxSize = attrs['BoxSize']

    # analytic randoms - randoms calculation assuming uniform distribution
    R1R2 = AnalyticUniformRandoms(mode, D1D2.pairs.dims, D1D2.pairs.edges, BoxSize)(ND1, ND2)

    # and compute the correlation function as DD/RR - 1
    fDD = R1R2.attrs['total_wnpairs'] / D1D2.attrs['total_wnpairs']
    RR = R1R2.pairs['wnpairs']
    DD = D1D2.pairs['wnpairs']

    CF = (DD * fDD) / RR - 1.

    # create a BinnedStatistic holding the CF
    CF = _create_tpcf_result(D1D2.pairs, CF)

    return R1R2.pairs, CF

def _create_tpcf_result(D1D2, CF):
    """
    Create a BinnedStatistic holding the correlation function
    and average bin separation.
    """
    x = D1D2.dims[0]
    data = numpy.empty_like(CF, dtype=[('corr', 'f8'), (x, 'f8')])
    data['corr'] = CF[:]
    data[x] = D1D2[x]
    edges = [D1D2.edges[d] for d in D1D2.dims]
    return WedgeBinnedStatistic(D1D2.dims, edges, data)



from nbodykit.cosmology import Cosmology, Planck15, WMAP9
from numpy.testing import assert_allclose, assert_array_equal
import numpy
import pytest

def test_old_Omega_syntax():

    c1 = Cosmology(Omega_b=0.04)
    c2 = Cosmology(Omega0_b=0.04)
    assert c1.Omega0_b == c2.Omega0_b

    c1 = Cosmology(T_cmb=2.7)
    c2 = Cosmology(T0_cmb=2.7)
    assert c1.T0_cmb == c2.T0_cmb

    c1 = Cosmology(Omega0_k=0.05)
    c2 = Cosmology(Omega_k=0.05)
    assert c1.Omega0_k == c2.Omega0_k

    c1 = Cosmology(Omega0_lambda=0.7)
    c2 = Cosmology(Omega_lambda=0.7)
    c3 = Cosmology(Omega0_Lambda=0.7)
    assert c1.Omega0_lambda == c2.Omega0_lambda
    assert c1.Omega0_lambda == c3.Omega0_lambda

def test_deprecated_init():

    # all valid deprecated kwargs
    with pytest.warns(FutureWarning):
        c1 = Cosmology(H0=67.6, Om0=0.31, flat=True)
        c2 = Cosmology(H0=67.6, Om0=0.31, Ode0=0.7, flat=False, w0=-0.9)

    # missing valid and deprecated
    with pytest.raises(Exception):
        c = Cosmology(h=0.7, flat=True)

    # passing arguments and mixing
    with pytest.raises(Exception):
        c = Cosmology(0.7, flat=True)

    # parameter conflict
    with pytest.raises(Exception):
        c3 = Cosmology(H0=70., flat=True, h=0.7)

    assert_allclose(c1.h, 0.676)
    assert_allclose(c2.h, 0.676)
    assert_allclose(c1.Om0, 0.31)
    assert_allclose(c2.Om0, 0.31)
    assert_allclose(c1.Ok0, 0.)
    assert_allclose(c2.Ode0, 0.7)
    assert_allclose(c2.w0_fld, -0.9)

def test_efunc_prime():
    epsilon = 1e-4
    z = numpy.linspace(0, 3, 1000) + epsilon

    # cosmology with no massive neutrinos
    cosmo = WMAP9
    d1 = cosmo.efunc_prime(z)
    d2 = (cosmo.efunc(z+epsilon) - cosmo.efunc(z-epsilon))/(2*epsilon) * -(1+z)**2
    assert_allclose(d1, d2, err_msg="efunc_prime error for cosmo with no massive neutrinos", rtol=1e-3)

    # cosmology with massive neutrinos
    cosmo = Planck15
    d1 = cosmo.efunc_prime(z)
    d2 = (cosmo.efunc(z+epsilon) - cosmo.efunc(z-epsilon))/(2*epsilon) * -(1+z)**2
    assert_allclose(d1, d2, err_msg="efunc_prime error for cosmo with massive neutrinos", rtol=1e-3)

def test_load_precision():
    from classylss import load_precision

    p = load_precision('pk_ref.pre')
    c = Cosmology(gauge='synchronous', tol_background_integration=1e-5, **p)
    assert_allclose(c.Omega_cdm(0), c.Omega0_cdm)

def test_clone():
    c = Cosmology(gauge='synchronous', tol_background_integration=1e-5)
    c2 = c.clone(Omega0_b=0.04)
    assert_allclose(c2.Omega0_b, 0.04)
    c2 = c2.clone()
    assert_allclose(c2.Omega0_b, 0.04)

def test_from_file():

    import tempfile, pickle
    with tempfile.NamedTemporaryFile(mode='w') as ff:
        ff.write("H0=70\nomega_b = 0.0266691\nomega_cdm = 0.110616\nT_cmb=2.7255\n")
        ff.seek(0)

        # load from a file and check values
        c = Cosmology.from_file(ff.name)
        assert_allclose(c.Omega0_b*c.h**2, 0.0266691)
        assert_allclose(c.Omega0_cdm*c.h**2, 0.110616)

    # clone
    c2 = c.clone(Omega0_b=0.04)
    assert_allclose(c2.Omega0_b, 0.04)

    # serialize and make sure we get the same
    s = pickle.dumps(c)
    c1 = pickle.loads(s)

    assert_allclose(c.Omega0_cdm, c1.Omega0_cdm)
    assert_allclose(c.Omega0_b, c1.Omega0_b)

def test_conflicts():

    # h is the correct param
    with pytest.raises(Exception):
        c = Cosmology(h=0.7, H0=70)

    with pytest.raises(Exception):
        c = Cosmology(Omega0_b=0.04, Omega_b=0.04)

    # Omega_b is the correct param
    with pytest.raises(Exception):
        c = Cosmology(Omega0_b=0.04, omega_b=0.02)

def test_unknown_params():

    # warn about unknown parameters
    with pytest.warns(UserWarning):
        c = Cosmology(unknown_paramter=100.)

def test_set_sigma8():

    # set sigma8 by adjusting A_s internally
    c = Cosmology().match(sigma8=0.80)

    # run CLASS and compute sigma8
    assert_allclose(c.sigma8, 0.80)

def test_set_Omega0_cb():

    c = Cosmology().match(Omega0_cb=0.4)

    assert_allclose(c.Omega0_cb, 0.4)

    c = Cosmology().match(Omega0_m=0.4)
    assert_allclose(c.Omega0_m, 0.4)


def test_sigma8_z():

    z = numpy.linspace(0, 1, 100)
    c = Cosmology()

    s8_z = c.sigma8_z(z)
    D_z = c.scale_independent_growth_factor(z)
    assert_allclose(s8_z, D_z*c.sigma8, rtol=1e-3)


def test_cosmology_sane():
    c = Cosmology(gauge='synchronous', verbose=True)

    assert_allclose(c.Omega_cdm(0), c.Omega0_cdm)
    assert_allclose(c.Omega_g(0), c.Omega0_g)
    assert_allclose(c.Omega_b(0), c.Omega0_b)
    assert_allclose(c.Omega_ncdm(0), c.Omega0_ncdm)
    assert_allclose(c.Omega_ur(0), c.Omega0_ur)
    assert_allclose(c.Omega_ncdm(0), c.Omega0_ncdm_tot)

    assert_allclose(c.Omega_pncdm(0), c.Omega0_pncdm)
    assert_allclose(c.Omega_m(0), c.Omega0_m)
    assert_allclose(c.Omega_r(0), c.Omega0_r)

    # total density in 10e10 Msun/h unit
    assert_allclose(c.rho_tot(0), 27.754999)

    # comoving distance to z=1.0 in Mpc/h unit.
    assert_allclose(c.comoving_distance(1.0), 3396.157391 * c.h)

    # conformal time in Mpc unit.
    assert_allclose(c.tau(1.0), 3396.157391)

    assert_allclose(c.efunc(0), 1.) # hubble in Mpc/h km/s unit
    assert_allclose(c.efunc(0) - c.efunc(1 / 0.9999 - 1),
                    0.0001 * c.efunc_prime(0), rtol=1e-3)

def test_cosmology_density():
    c = Cosmology(gauge='synchronous')
    z = [0, 1, 2, 5, 9, 99]
    assert_allclose(c.rho_cdm(z), c.Omega_cdm(z) * c.rho_tot(z))
    assert_allclose(c.rho_g(z), c.Omega_g(z) * c.rho_tot(z))
    assert_allclose(c.rho_ncdm(z), c.Omega_ncdm(z) * c.rho_tot(z))
    assert_allclose(c.rho_b(z), c.Omega_b(z) * c.rho_tot(z))
    assert_allclose(c.rho_m(z), c.Omega_m(z) * c.rho_tot(z))
    assert_allclose(c.rho_r(z), c.Omega_r(z) * c.rho_tot(z))
    assert_allclose(c.rho_ur(z), c.Omega_ur(z) * c.rho_tot(z))

def test_cosmology_vect():
    c = Cosmology(gauge='synchronous')

    assert_allclose(c.Omega_cdm([0]), c.Omega0_cdm)

    assert_array_equal(c.Omega_cdm([]).shape, [0])
    assert_array_equal(c.Omega_cdm([0]).shape, [1])
    assert_array_equal(c.Omega_cdm([[0]]).shape, [1, 1])

    assert_array_equal(c.rho_k([[0]]).shape, [1, 1])

    k, z = numpy.meshgrid([0, 1], [0.01, 0.05, 0.1, 0.5], sparse=True, indexing='ij')

    pk = c.get_pk(z=z, k=k)
    assert_array_equal(pk.shape, [2, 4])

def test_cosmology_a_max():
    c = Cosmology(gauge='synchronous', a_max=2.0)
    print(c.parameter_file)
    assert c.a_max == 2.0
    t = c.Omega_m(-0.1)
    t = c.efunc(-0.1)
    t = c.scale_independent_growth_factor(-0.1)

    #t = c.get_transfer(z=-0.1)

def test_cosmology_transfer():
    c = Cosmology()
    t = c.get_transfer(z=0)
    assert 'h_prime' in t.dtype.names
    assert 'k' in t.dtype.names
    assert 'd_cdm' in t.dtype.names

def test_cosmology_get_pk():
    c = Cosmology()
    p = c.get_pk(z=0, k=0.1)
    p1 = c.Spectra.get_pk(z=0, k=0.1)

    # ensure the dro did use get_pk of Spectra rather than that from Primordial
    assert_allclose(p, p1)

def test_to_astropy():

    from astropy.cosmology import FlatLambdaCDM, LambdaCDM
    from astropy.cosmology import FlatwCDM, wCDM
    from astropy.cosmology import Flatw0waCDM, w0waCDM

    # lambda CDM
    for cls in [FlatLambdaCDM, LambdaCDM]:
        if "Flat" in cls.__name__:
            c1 = Cosmology(Omega_k=0.)
        else:
            c1 = Cosmology(Omega_k=0.05)
        c2 = c1.to_astropy()
        assert isinstance(c2, cls)
        assert_allclose(c2.Ok0, c1.Omega0_k, rtol=1e-3)

    # w0 CDM
    for cls in [FlatwCDM, wCDM]:
        if "Flat" in cls.__name__:
            c1 = Cosmology(w0_fld=-0.9, Omega_k=0.)
        else:
            c1 = Cosmology(w0_fld=-0.9, Omega_k=0.05)
        c2 = c1.to_astropy()
        assert isinstance(c2, cls)
        assert_allclose(c2.Ok0, c1.Omega0_k, rtol=1e-3)
        assert_allclose(c2.w0, c1.w0_fld)

    # wa w0 CDM
    for cls in [Flatw0waCDM, w0waCDM]:
        if "Flat" in cls.__name__:
            c1 = Cosmology(wa_fld=0.05, w0_fld=-0.9, Omega_k=0.)
        else:
            c1 = Cosmology(wa_fld=0.05, w0_fld=-0.9, Omega_k=0.05)
        c2 = c1.to_astropy()
        assert isinstance(c2, cls)
        assert_allclose(c2.Ok0, c1.Omega0_k, rtol=1e-3)
        assert_allclose(c2.w0, c1.w0_fld)
        assert_allclose(c2.wa, c1.wa_fld)

def test_from_astropy():

    from astropy.cosmology import FlatLambdaCDM, LambdaCDM
    from astropy.cosmology import FlatwCDM, wCDM
    from astropy.cosmology import Flatw0waCDM, w0waCDM

    # LambdaCDM
    flat = {'H0':70, 'Om0':0.3, 'Ob0':0.04, 'Tcmb0':2.7}
    for cls in [FlatLambdaCDM, LambdaCDM]:
        if "Flat" in cls.__name__:
            x = cls(**flat)
        else:
            x = cls(Ode0=0.75, **flat)
        c = Cosmology.from_astropy(x)
        assert_allclose(c.Ok0, x.Ok0)
        assert_allclose(c.Omega0_fld, 0.) # Omega0_lambda is nonzero
        assert_allclose(c.Odm0, x.Odm0)

    # w0 CDM
    for cls in [FlatwCDM, wCDM]:
        if "Flat" in cls.__name__:
            x = cls(w0=-0.9, **flat)
        else:
            x = cls(w0=-0.9, Ode0=0.75, **flat)
        c = Cosmology.from_astropy(x)
        assert_allclose(c.Ok0, x.Ok0)
        assert_allclose(c.Odm0, x.Odm0)
        assert_allclose(c.w0, x.w0)
        assert_allclose(c.Omega0_lambda, 0.) # Omega_fld is nonzero

    # w0,wa CDM
    for cls in [Flatw0waCDM, w0waCDM]:
        if "Flat" in cls.__name__:
            x = cls(w0=-0.9, wa=0.01, **flat)
        else:
            x = cls(w0=-0.9, wa=0.01, Ode0=0.75, **flat)
        c = Cosmology.from_astropy(x)
        assert_allclose(c.Ok0, x.Ok0)
        assert_allclose(c.Odm0, x.Odm0)
        assert_allclose(c.w0, x.w0)
        assert_allclose(c.wa, x.wa)
        assert_allclose(c.Omega0_lambda, 0.) # Omega_fld is nonzero

def test_immutable():

    c = Cosmology()
    with pytest.raises(ValueError):
        c.A_s = 2e-9 # immutable

    # can add non-CLASS attributes still
    c.test = 'TEST'

def test_bad_no_Ob0():

    from astropy.cosmology import FlatLambdaCDM
    c = FlatLambdaCDM(Om0=0.3, H0=70) # no Ob0

    with pytest.raises(ValueError):
        c = Cosmology.from_astropy(c)

def test_bad_astropy_class():

    from astropy.cosmology import w0wzCDM
    c = w0wzCDM(Om0=0.3, H0=70, Ode0=0.7) # no Ob0

    with pytest.raises(ValueError):
        c = Cosmology.from_astropy(c)

def test_massive_neutrinos():

    # single massive neutrino
    c = Cosmology(m_ncdm=0.6)
    assert c.N_ncdm == 1

    # do not need 0 values
    with pytest.raises(ValueError):
        c = Cosmology(m_ncdm=[0.6, 0.])

def test_no_massive_neutrinos():

    c = Cosmology(m_ncdm=None)
    assert c.has_massive_nu == False

def test_bad_input():

    with pytest.raises(ValueError):
        c = Cosmology(gauge='BAD')

    # specifying w0_fld + Omega_Lambda is inconsistent
    with pytest.raises(ValueError):
        c = Cosmology(Omega_Lambda=0.7, w0_fld=-0.9)


def test_cosmology_dir():
    c = Cosmology()
    d = dir(c)
    assert "Background" in d
    assert "Spectra" in d
    assert "Omega0_m" in d

def test_cosmology_pickle():
    import pickle
    c = Cosmology()
    s = pickle.dumps(c)
    c1 = pickle.loads(s)
    assert c1.parameter_file == c.parameter_file

def test_cosmology_clone():
    c = Cosmology(gauge='synchronous')

    c1 = Cosmology(gauge='newtonian')
    assert 'newtonian' in c1.parameter_file

    c2 = Cosmology(P_k_max=1.01234567)
    assert '1.01234567' in c2.parameter_file

def test_astropy_compat():
    c = Cosmology(gauge='synchronous', m_ncdm=[0.06])

    assert_allclose(c.Odm(0), c.Odm0)
    assert_allclose(c.Ogamma(0), c.Ogamma0)
    assert_allclose(c.Ob(0), c.Ob0)
    assert_allclose(c.Onu(0), c.Onu0)
    assert_allclose(c.Ok(0), c.Ok0)
    assert_allclose(c.Ode(0), c.Ode0)
    assert_array_equal(c.has_massive_nu, True)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from nbodykit.utils import GatherArray

from numpy.testing import assert_allclose
from numpy.testing import assert_array_equal
from mpi4py import MPI

setup_logging("debug")

@MPITest([4])
def test_lognormal_sparse(comm):
    cosmo = cosmology.Planck15

    # this should generate 15 particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=1e-5, BoxSize=128., Nmesh=8, seed=42, comm=comm)

    mesh = source.to_mesh(compensated=False)

    real = mesh.compute(mode='real')
    assert_allclose(real.cmean(), 1.0, 1e-5)

@MPITest([1, 4])
def test_lognormal_dense(comm):
    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=0.2e-2, BoxSize=128., Nmesh=8, seed=42, comm=comm)
    mesh = source.to_mesh(compensated=False)

    real = mesh.compute(mode='real')
    assert_allclose(real.cmean(), 1.0, rtol=1e-5)

@MPITest([4])
def test_lognormal_invariance(comm):
    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=0.5e-2, BoxSize=128., Nmesh=32, seed=42, comm=comm)
    source1 = LogNormalCatalog(Plin=Plin, nbar=0.5e-2, BoxSize=128., Nmesh=32, seed=42, comm=MPI.COMM_SELF)

    assert source.csize == source1.size

    allpos = GatherArray(source['Position'].compute(), root=Ellipsis, comm=comm)
    assert_allclose(allpos, source1['Position'])
    alldis = GatherArray(source['Velocity'].compute(), root=Ellipsis, comm=comm)
    assert_allclose(alldis, source1['Velocity'])

@MPITest([1])
def test_lognormal_velocity(comm):
    cosmo = cosmology.Planck15

    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=0.5e-2, BoxSize=128., Nmesh=32, seed=42, comm=comm)

    source['Value'] = source['Velocity'][:, 0]**2
    mesh = source.to_mesh(compensated=False)

    real = mesh.compute(mode='real')
    velsum = comm.allreduce((source['Velocity'][:, 0]**2).sum().compute())
    velmean = velsum / source.csize

    assert_allclose(real.cmean(), velmean, rtol=1e-5)
